#include <fcntl.h>
#include <inttypes.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/mman.h>
#include <sys/stat.h>
#include <sys/types.h>

#define handle_error(msg) \
    do { perror(msg); exit(EXIT_FAILURE); } while (0)

int
main(int argc, char *argv[])
{
    int fd, a;
    struct stat sb;
    size_t len;
    struct pstack {             /* persistent stack header */
        int32_t n;              /* number of items in stack */
        int32_t s[];            /* array containing stack */
    } *p;                       /* ptr to mmap'd stack file */

    fd = open("pstack", O_RDWR);
    if (fd == -1)
        handle_error("open");

    if (fstat(fd, &sb) != 0)    /* to obtain file size */
        handle_error("fstat");

    len = (size_t) sb.st_size;

    p = (struct pstack *) mmap(NULL, len, PROT_READ | PROT_WRITE,
                               MAP_SHARED, fd, 0);
    if (p == MAP_FAILED)
        handle_error("mmap");

    for (a = 1; a < argc; a++) {
        if (strcmp(argv[a], "pop") == 0) {
            if (p->n <= 0) {
                printf("<stack empty>\n");
                exit(EXIT_FAILURE);
            } else {
                printf("%" PRId32 "\n", p->s[--p->n]);
            }
        } else {
            if (sizeof *p + (size_t) p->n * sizeof p->n >= len) {
                printf("<stack full>\n");
                exit(EXIT_FAILURE);
            } else {
                p->s[p->n++] = atoi(argv[a]);
            }
        }
    }

    if (close(fd) != 0)         /* implicit on exit() */
        handle_error("close");

    if (munmap(p, len) != 0)    /* implicit on exit() */
        handle_error("munmap");

    exit(EXIT_SUCCESS);
}
